/* ----------------------------------------------------------------------------
 * rishu_proxy.cc
 *   里珠/Proxy の栞のみでの実装.
 * ----------------------------------------------------------------------------
 * Mastering programed by YAMASHINA Hio
 *
 * Copyright 2002 YMIRLINK,Inc.
 * ----------------------------------------------------------------------------
 * $Id: rishu_proxy.cc,v 1.4 2002/12/21 09:28:59 hio Exp $
 * ----------------------------------------------------------------------------
 */

#ifndef _WIN32
#error written for win32 only
#endif
#include <windows.h>
#include "rishu_proxy.h"
#include "resource.h"
#include <stdio.h>
#include <string.h>
#include <tchar.h>
#include <winsock.h>

/* ----------------------------------------------------------------------------
 * LPTSTR message = getLastErrorMessage();
 * LPTSTR message = getErrorMessage(DWORD errorCode);
 *   エラーメッセージの取得 
 *   取得したメッセージは LocalFree で解放してね☆ 
 */
LPTSTR getLastErrorMessage(void)
{
  return getErrorMessage(GetLastError());
}
LPTSTR getErrorMessage(DWORD errcode)
{
  LPVOID lpMessage;
  DWORD msglen;
  lpMessage = NULL;
  msglen = FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER
			  | FORMAT_MESSAGE_FROM_SYSTEM
			  | FORMAT_MESSAGE_IGNORE_INSERTS,
			  NULL,
			  errcode,
			  MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT), /* 既定の言語 */
			  (LPTSTR)&lpMessage,
			  0,
			  NULL
			  );
  if( msglen==0 )
  {
    if( lpMessage )
    {
      lpMessage = LocalReAlloc(lpMessage,64,0);
    }else
    {
      lpMessage = LocalAlloc(LMEM_FIXED,64);
    }
    if( lpMessage )
    {
      _sntprintf((LPTSTR)lpMessage,64,
		 TEXT("Unknown Error (%lu,0x%08x)\n"),
		 errcode, errcode
		 );
    }
  }
  return lpMessage;
}

/* ----------------------------------------------------------------------------
 * ANSI => Unicode 変換
 *   result: 書き込まれた文字数(TCHARの数).
 *           エラー時 0.
 *           nOutStringLen==0 だと必要な文字数.
 */
int
AnsiToUnicode(LPWSTR lpszOutputString,
	      int    nOutStringLen,
	      LPCSTR lpInputString,
	      int    nInStringLen
	      )
{
  return MultiByteToWideChar(CP_ACP,
			     0,
			     lpInputString, nInStringLen,
			     lpszOutputString, nOutStringLen);
}

/* ----------------------------------------------------------------------------
 * Unicode => ANSI 変換.
 */
int
UnicodeToAnsi(LPSTR   lpszOutputString,
	      int     nOutStringLen,
	      LPCWSTR lpInputString,
	      int     nInStringLen
	      )
{
  return WideCharToMultiByte(CP_ACP,
			     0,
			     lpInputString, nInStringLen,
			     lpszOutputString, nOutStringLen,
			     NULL, NULL
			     );
}


/* ----------------------------------------------------------------------------
 * 里珠/Proxy の初期化.
 */
RishuProxy* RishuProxy_init(RishuProxy* proxy,
			    WORD wSize,
			    HANDLE hDll
			    )
{
  if( wSize<sizeof(RishuProxy) )
  {
    return NULL;
  }
  
  /* 初期値の設定 */
  proxy->wSize = wSize;
  proxy->wVersion = 0x0101;
  
  proxy->hDll = hDll;
  proxy->hGlobal_loaddir = NULL;
  proxy->loaddir = NULL;
  proxy->dwLoaddirLen = 0;
  
  proxy->hChlProcess = NULL;
#ifdef RISHU_PROXY_USESOCKET
  proxy->communicationType = RP_CT_TCP;
#else
  proxy->communicationType = RP_CT_PIPE;
#endif
  proxy->hChlErrorOutput = NULL;
#ifdef RISHU_PROXY_USEPIPE
  proxy->hChlInput = NULL;
  proxy->hChlOutput = NULL;
#endif
#ifdef RISHU_PROXY_USESOCKET
  proxy->sockListen = INVALID_SOCKET;
  proxy->sockRemote = INVALID_SOCKET;
  proxy->tvTimeout.tv_sec = 300;
  proxy->tvTimeout.tv_usec = 0;
#endif
  
  proxy->dwLogBits =
    (1<<RishuProxy_LOG_FATAL)
    | (1<<RishuProxy_LOG_ERROR)
    | (1<<RishuProxy_LOG_SYSTEM)
    | (1<<RishuProxy_LOG_CLEANUP)
    ;
  proxy->logfilename = NULL;
  proxy->fp = NULL;
  
  proxy->msgseq = 0;
  proxy->readbuf = NULL;
  proxy->dwReadBufLen = 0;
  proxy->dwReadBufSize = 0;
  
  proxy->hHeap = NULL;
  
  /* あれこれオブジェクトの初期化. */
  
  /* ヒープオブジェクトの作成 */
  proxy->hHeap = HeapCreate(HEAP_NO_SERIALIZE,64*1024,0);
  if( proxy->hHeap==NULL )
  {
    return NULL;
  }
  
  /* ini ファイルの確認. */
  {
    LPTSTR filename;
    DWORD  filename_charsize;
    DWORD  len;
    DWORD  tail;
    
    DWORD dwRes;
    LPTSTR tmpbuf;
    DWORD tmpbuf_charsize;
    DWORD i;
      
    /* ファイル名の生成 */
    filename_charsize = 512;
    filename = (LPTSTR)proxy->async_myAlloc(filename_charsize*sizeof(TCHAR));
    if( !filename )
    {
      proxy->discard();
      return NULL;
    }
    filename_charsize = proxy->async_mySize(filename);
    for(;;)
    {
      len = GetModuleFileName(proxy->hDll,filename,filename_charsize-1);
      if( len==0 )
      {
	/* GetModuleFileName failed. */
	proxy->async_myFree(filename);
	proxy->discard();
	return NULL;
      }
      if( len<filename_charsize-2 )
      {
	break;
      }
      filename_charsize *= 2;
      filename = (LPTSTR)proxy->async_myReAlloc(filename,filename_charsize*sizeof(TCHAR));
      if( !filename )
      {
	proxy->discard();
	return NULL;
      }
      filename_charsize = proxy->async_mySize(filename);
    }
    
    /* 拡張子辺りの書き換え */
    tail = len;
    for(;;)
    {
      if( tail==0 || filename[--tail]==TEXT('\\') )
      {
	tail = len;
	break;
      }
      if( filename[tail]==TEXT('.') )
      {
	break;
      }
    }
    if( tail+4+1>=filename_charsize )
    {
      filename_charsize = tail+4+1;
      filename = (LPTSTR)proxy->async_myReAlloc(filename,filename_charsize*sizeof(TCHAR));
      if( !filename )
      {
	proxy->discard();
	return NULL;
      }
      filename_charsize = proxy->async_mySize(filename);
    }
    _tcsncpy(filename+tail,TEXT(".ini"),5);
    
    /* ログファイル名も生成〜. */
    proxy->logfilename  = (LPTSTR)proxy->async_myAlloc(tail+9);
    if( !proxy->logfilename )
    {
      proxy->discard();
      return NULL;
    }
    CopyMemory(proxy->logfilename,filename,tail*sizeof(TCHAR));
    CopyMemory(proxy->logfilename+tail,TEXT("_log.txt"),9*sizeof(TCHAR));
    proxy->log(TEXT("system"),
		   TEXT("RishuProxy::init"),
		   TEXT("now initializing ...\n")
		   );
    
    /* 設定の読み込み. */
    tmpbuf_charsize = 512;
    tmpbuf = (LPTSTR)proxy->async_myAlloc(tmpbuf_charsize*sizeof(TCHAR));

    /* ログ設定 */
    proxy->log(TEXT("system"),
		   TEXT("RishuProxy::init"),
		   TEXT("loading ini file ... [%s]\n"),
		   filename
		   );
    for(;;)
    {
      dwRes = GetPrivateProfileString(TEXT("Rishu/Proxy"),
				      TEXT("log"),
				      TEXT("nya"),
				      tmpbuf,
				      tmpbuf_charsize,
				      filename
				      );
      if( dwRes!=tmpbuf_charsize-1 )
      {
	break;
      }
      tmpbuf_charsize *= 2;
      tmpbuf = (LPTSTR)proxy->async_myAlloc(tmpbuf_charsize*sizeof(TCHAR));
      if( !tmpbuf )
      {
	proxy->discard();
	return NULL;
      }
    }
    
    for( i=0; i<dwRes; ++i )
    {
      DWORD st;
      BOOL  on;
      DWORD bit;
      if( _istspace(tmpbuf[i]) )
      {
	continue;
      }
      if( tmpbuf[i]==TEXT('-') )
      {
	on = FALSE;
	++i;
      }else if( tmpbuf[i]==TEXT('+') )
      {
	on = TRUE;
	++i;
      }else
      {
	on = TRUE;
      }
      st = i;
      while( ++i<dwRes && !_istspace(tmpbuf[i]) )
      {
      }
      if( i-st==5 && _tcsnicmp(tmpbuf+st,TEXT("fatal"),5)==0 )
      {
	bit = 1<<RishuProxy_LOG_FATAL;
      }else if( i-st==5 && _tcsnicmp(tmpbuf+st,TEXT("error"),5)==0 )
      {
	bit = 1<<RishuProxy_LOG_ERROR;
      }else if( i-st==4 && _tcsnicmp(tmpbuf+st,TEXT("warn"),4)==0 )
      {
	bit = 1<<RishuProxy_LOG_WARN;
      }else if( i-st==6 && _tcsnicmp(tmpbuf+st,TEXT("system"),6)==0 )
      {
	bit = 1<<RishuProxy_LOG_SYSTEM;
	on = TRUE;
      }else if( i-st==5 && _tcsnicmp(tmpbuf+st,TEXT("trace"),5)==0 )
      {
	bit = 1<<RishuProxy_LOG_TRACE;
      }else if( i-st==5 && _tcsnicmp(tmpbuf+st,TEXT("debug"),5)==0 )
      {
	bit = 1<<RishuProxy_LOG_DEBUG;
      }else if( i-st==6 && _tcsnicmp(tmpbuf+st,TEXT("stderr"),6)==0 )
      {
	bit = 1<<RishuProxy_LOG_STDERR;
      }else
      {
	proxy->log(TEXT("system"),
		       TEXT("RishuProxy::init"),
		       TEXT("[unknown logtype] [[%.*s]]\n"),
		       i-st, tmpbuf+st
		       );
	continue;
      }
      if( on )
      {
	proxy->dwLogBits |= bit;
      }else
      {
	proxy->dwLogBits &= ~bit;
      }
    }
  }
  proxy->dwLogBits |= 1<<RishuProxy_LOG_SYSTEM;
  if( proxy->tracelogIsEnabled() )
  {
    proxy->log(TEXT("trace"),
		   TEXT("RishuProxy::init"),
		   TEXT("initialize done\n"
			"[[\nlogbits==%08x\n]]\n"),
		   proxy->dwLogBits
		   );
  }
  
  return proxy;
}

/* ----------------------------------------------------------------------------
 * 里珠/Proxy の破棄.
 */
void RishuProxy_discard(RishuProxy* proxy)
{
  if( proxy->loaddir )
  {
    if( proxy->loaddir!=(LPTSTR)proxy->hGlobal_loaddir )
    {
      proxy->myFree(proxy->loaddir);
    }
    proxy->loaddir = NULL;
    proxy->dwLoaddirLen = 0;
  }
  if( proxy->hGlobal_loaddir )
  {
    GlobalFree(proxy->hGlobal_loaddir);
    proxy->hGlobal_loaddir = NULL;
    proxy->dwLoaddirLen = 0;
  }
  if( proxy->hChlProcess )
  {
    TerminateProcess(proxy->hChlProcess,1);
    CloseHandle(proxy->hChlProcess);
    proxy->hChlProcess = NULL;
  }
  if( proxy->hChlErrorOutput )
  {
    CloseHandle(proxy->hChlErrorOutput);
    proxy->hChlErrorOutput = NULL;
  }
#ifdef RISHU_PROXY_USEPIPE
  if( proxy->hChlInput )
  {
    CloseHandle(proxy->hChlInput);
    proxy->hChlInput = NULL;
  }
  if( proxy->hChlOutput )
  {
    CloseHandle(proxy->hChlOutput);
    proxy->hChlOutput = NULL;
  }
#endif
#ifdef RISHU_PROXY_USESOCKET
  if( proxy->sockListen!=INVALID_SOCKET )
  {
    closesocket(proxy->sockListen);
    proxy->sockListen = INVALID_SOCKET;
  }
  if( proxy->sockRemote!=INVALID_SOCKET )
  {
    closesocket(proxy->sockRemote);
    proxy->sockRemote = INVALID_SOCKET;
  }
#endif
  if( proxy->readbuf )
  {
    proxy->async_myFree(proxy->readbuf);
    proxy->readbuf = NULL;
    proxy->dwReadBufLen = 0;
    proxy->dwReadBufSize = 0;
  }
  if( proxy->fp )
  {
    fclose(proxy->fp);
    proxy->fp = NULL;
  }
  if( proxy->hHeap )
  {
    BOOL res = HeapDestroy(proxy->hHeap);
    proxy->hHeap = NULL;
  }
}

/* ----------------------------------------------------------------------------
 * 排他制御 - lock
 */
void RishuProxy_lock(RishuProxy* proxy)
{
  /* 未実装, not implemented yet */
  return;
}

/* ----------------------------------------------------------------------------
 * 排他制御 - trylock
 */
BOOL RishuProxy_trylock(RishuProxy* proxy)
{
  /* 未実装, not implemented yet */
  return TRUE;
}

/* ----------------------------------------------------------------------------
 * 排他制御 - unlock
 */
void RishuProxy_unlock(RishuProxy* proxy)
{
  /* 未実装, not implemented yet */
  return;
}

/* ---------------------------------------------------------------------------
 * ヒープ操作 - alloc
 */
LPVOID RishuProxy_async_myAlloc(RishuProxy* proxy,
				SIZE_T size
				)
{
  LPVOID ptr;
  
  if( !proxy )
  {
    return NULL;
  }
  
  /* メモリの確保 
   *   (失敗しても SetLastError はされない)
   */
  ptr = HeapAlloc(proxy->hHeap, HEAP_NO_SERIALIZE, size);

  return ptr;
}

/* ---------------------------------------------------------------------------
 * ヒープ操作 - realloc
 */
LPVOID RishuProxy_async_myReAlloc(RishuProxy* proxy,
				  LPVOID ptr,
				  DWORD size
				  )
{
  if( !proxy )
  {
    return NULL;
  }
  if( !proxy->hHeap )
  {
    if( proxy->fatallogIsEnabled() )
    {
      proxy->log(TEXT("fatal"),
		     TEXT("RishuProxy::async_myReAlloc"),
		     TEXT("!! no heap created yet !!\n")
		     );
    }
    return NULL;
  }
  
  return HeapReAlloc(proxy->hHeap,HEAP_NO_SERIALIZE,ptr,size);
}

/* ---------------------------------------------------------------------------
 * ヒープ操作 - free
 */
void RishuProxy_async_myFree(RishuProxy* proxy,
			     LPVOID ptr
			     )
{
  if( !proxy )
  {
    return;
  }
  if( !proxy->hHeap )
  {
    if( proxy->fatallogIsEnabled() )
    {
      proxy->log(TEXT("fatal"),
		     TEXT("RishuProxy::async_myFree"),
		     TEXT("!! no heap created yet !!\n")
		     );
    }
    return;
  }
  HeapFree(proxy->hHeap,HEAP_NO_SERIALIZE,ptr);
  return;
}

/* ---------------------------------------------------------------------------
 * ヒープ操作 - size
 */
SIZE_T RishuProxy_async_mySize(RishuProxy* proxy,
			       LPVOID ptr
			       )
{
  if( !proxy )
  {
    return 0;
  }
  if( !proxy->hHeap )
  {
    if( proxy->fatallogIsEnabled() )
    {
      proxy->log(TEXT("fatal"),
		     TEXT("RishuProxy::async_mySize"),
		     TEXT("!! no heap created yet !!\n")
		     );
    }
    return 0;
  }
  return HeapSize(proxy->hHeap,HEAP_NO_SERIALIZE,ptr);
}

/* ---------------------------------------------------------------------------
 * 同期付きヒープ操作 - alloc
 */
LPVOID RishuProxy_myAlloc(RishuProxy* proxy,
			  SIZE_T size
			  )
{
  LPVOID ptr;
  proxy->lock();
  ptr = proxy->async_myAlloc(size);
  proxy->unlock();
  return ptr;
}

/* ---------------------------------------------------------------------------
 * 同期付きヒープ操作 - realloc
 */
LPVOID RishuProxy_myReAlloc(RishuProxy* proxy,
			    LPVOID ptr,
			    SIZE_T size
			    )
{
  proxy->lock();
  ptr = proxy->async_myReAlloc(ptr,size);
  proxy->unlock();
  return ptr;
}

/* ---------------------------------------------------------------------------
 * 同期付きヒープ操作 - free
 */
void RishuProxy_myFree(RishuProxy* proxy,
		       LPVOID ptr
		       )
{
  proxy->lock();
  proxy->async_myFree(ptr);
  proxy->unlock();
  return;
}

/* ---------------------------------------------------------------------------
 * 同期付きヒープ操作 - size
 */
extern SIZE_T RishuProxy_mySize(RishuProxy* proxy,
				LPVOID ptr
				)
{
  SIZE_T size;
  proxy->lock();
  size = proxy->async_mySize(ptr);
  proxy->unlock();
  return size;
}


/* ----------------------------------------------------------------------------
 * loaddirからの相対ディレクトリを, loaddir を含めたものに変換.
 */
LPTSTR RishuProxy_getPath(RishuProxy* proxy,
			  LPCTSTR relativePath,
			  DWORD relativePath_len
			  )
{
  LPTSTR expanded;
  
  if( !proxy )
  {
    return NULL;
  }
  
  if( !proxy->loaddir )
  {
    if( !proxy->hGlobal_loaddir )
    {
      if( proxy->errorlogIsEnabled() )
      {
	proxy->log(TEXT("error"),
		       TEXT("RishuProxy::getPath"),
		       TEXT("on_load was not called yet.\n")
		       );
      }
    }
#ifndef UNICODE
    /* ANSI */
    proxy->loaddir = (LPSTR) proxy->hGlobal_loaddir;
#else
    /* UNICODE */
#error not suspported yet.
#endif /* defined(UNICODE) */
  }
  
  expanded = (LPTSTR)proxy->myAlloc((proxy->dwLoaddirLen+relativePath_len+1)*sizeof(TCHAR));
  if( !expanded )
  {
    if( proxy->errorlogIsEnabled() )
    {
      proxy->log(TEXT("error"),
		     TEXT("RishuProxy::getPath"),
		     TEXT("proxy->myAlloc failed\n")
		     );
    }
    return NULL;
  }
  CopyMemory(expanded,
	     proxy->loaddir,
	     proxy->dwLoaddirLen*sizeof(TCHAR)
	     );
  CopyMemory(expanded+proxy->dwLoaddirLen,
	     relativePath,
	     (relativePath_len+1)*sizeof(TCHAR)
	     );
  
  return expanded;
}


/* ----------------------------------------------------------------------------
 * 栞 Method / load
 */ 
BOOL RishuProxy_on_load(RishuProxy* proxy,
			HGLOBAL hGlobal_loaddir,
			long loaddir_len
			)
{
  BOOL res;
  
#ifdef RISHU_PROXY_USEPIPE
  HANDLE par_wd,par_rd;
  HANDLE chl_wd,chl_rd;
  HANDLE chl_wd_dup,chl_rd_dup;
#endif
  HANDLE hCurrentProcess;
  HANDLE err_rd,err_wd,err_wd_dup;

  LPTSTR commandLine;
  DWORD dwCreationFlags;
  STARTUPINFO startupInfo;
  PROCESS_INFORMATION processInfo;
  
  RishuProxy_Request request;
  RishuProxy_Response response;
  RishuProxy_Response* ptr_response;
  
  if( !proxy )
  {
    return FALSE;
  }

  if( proxy->tracelogIsEnabled() )
  {
    proxy->log(TEXT("trace"),
		   TEXT("RishuProxy::on_load"),
		   TEXT("enter\n")
		   );
  }
  
  /* --------------------------------------------------------------------------
   * 接続中のプロセスがあったら終了しておく (あるはずないけど) 
   */
  if( proxy->hChlProcess )
  {
    if( proxy->warnlogIsEnabled() )
    {
      proxy->log(TEXT("warn"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("remote process is already running with pid 0x%08x, do terminate\n"),
		     proxy->hChlProcess
		     );
    }
    res = TerminateProcess(proxy->hChlProcess,1);
    proxy->hChlProcess = NULL;
    if( !res && proxy->warnlogIsEnabled() )
    {
      LPTSTR cause = getLastErrorMessage();
      proxy->log(TEXT("warn"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("terminateProcess failed\n>> %s\n"),
		     cause
		     );
      LocalFree(cause);
    }
  }
#ifndef RISHU_PROXY_USEPIPE
  if( proxy->hChlInput )
  {
    CloseHandle(proxy->hChlInput);
    proxy->hChlInput = NULL;
  }
  if( proxy->hChlOutput )
  {
    CloseHandle(proxy->hChlOutput);
    proxy->hChlOutput = NULL;
  }
#endif
#ifdef RISHU_PROXY_USESOCKET
  if( proxy->sockListen!=INVALID_SOCKET )
  {
    closesocket(proxy->sockListen);
    proxy->sockListen = INVALID_SOCKET;
  }
  if( proxy->sockRemote!=INVALID_SOCKET )
  {
    closesocket(proxy->sockRemote);
    proxy->sockRemote = INVALID_SOCKET;
  }
#endif
  
  /* 古いデータがあったらけしておく. */
  if( proxy->loaddir )
  {
    if( proxy->loaddir!=(LPTSTR)proxy->hGlobal_loaddir )
    {
      proxy->myFree(proxy->loaddir);
    }
    proxy->loaddir = NULL;
  }
  if( proxy->hGlobal_loaddir )
  {
    GlobalFree(proxy->hGlobal_loaddir);
  }
  
  /* --------------------------------------------------------------------------
   * loaddir を格納.
   */
  proxy->hGlobal_loaddir = hGlobal_loaddir;
  proxy->dwLoaddirLen = loaddir_len;
    
  /* --------------------------------------------------------------------------
   * 通信用のパイプの作成.
   */
  hCurrentProcess = GetCurrentProcess();
  
  /* 標準エラーの作成. */
  if( !CreatePipe(&err_rd,&err_wd,NULL,0) )
  {
    if( proxy->errorlogIsEnabled() )
    {
      LPTSTR cause = getLastErrorMessage();
      proxy->log(TEXT("error"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("CreatePipe failed (for stderr)\n>> %s\n"),
		     cause
		     );
      LocalFree(cause);
    }
    if( proxy->tracelogIsEnabled() )
    {
      proxy->log(TEXT("trace"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("failure, CreatePipe (for stderr)\n")
		     );
    }
    return FALSE;
  }
  if( !DuplicateHandle(hCurrentProcess, err_wd,
		       hCurrentProcess, &err_wd_dup,
		       0,
		       TRUE,
		       DUPLICATE_SAME_ACCESS
		       ) )
  {
    if( proxy->errorlogIsEnabled() )
    {
      LPTSTR cause = getLastErrorMessage();
      proxy->log(TEXT("error"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("DuplicateHandle failed (for stderr)\n>> %s\n"),
		     cause
		     );
      LocalFree(cause);
    }
    if( proxy->tracelogIsEnabled() )
    {
      proxy->log(TEXT("trace"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("failure, DuplicateHandle (for stderr)\n")
		     );
    }
    CloseHandle(err_rd);
    CloseHandle(err_wd);
    return FALSE;
  }
  CloseHandle(err_wd);
  err_wd = NULL;
  
  if( proxy->communicationType==RP_CT_PIPE )
  {
  if( !CreatePipe(&par_rd,&chl_wd,NULL,0) )
  {
    if( proxy->errorlogIsEnabled() )
    {
      LPTSTR cause = getLastErrorMessage();
      proxy->log(TEXT("error"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("CreatePipe failed (1/2)\n>> %s\n"),
		     cause
		     );
      LocalFree(cause);
    }
    if( proxy->tracelogIsEnabled() )
    {
      proxy->log(TEXT("trace"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("failure, CreatePipe (1/2)\n")
		     );
    }
    return FALSE;
  }
  if( !DuplicateHandle(hCurrentProcess, chl_wd,
		       hCurrentProcess, &chl_wd_dup,
		       0,
		       TRUE,
		       DUPLICATE_SAME_ACCESS
		       ) )
  {
    if( proxy->errorlogIsEnabled() )
    {
      LPTSTR cause = getLastErrorMessage();
      proxy->log(TEXT("error"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("DuplicateHandle failed (1/2)\n>> %s\n"),
		     cause
		     );
      LocalFree(cause);
    }
    CloseHandle(par_rd);
    CloseHandle(chl_wd);
    if( proxy->tracelogIsEnabled() )
    {
      proxy->log(TEXT("trace"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("failure, DuplicateHandle (1/2)\n")
		     );
    }
    return FALSE;
  }
  CloseHandle(chl_wd);
  chl_wd = NULL;
  
  if( !CreatePipe(&chl_rd,&par_wd,NULL,0) )
  {
    if( proxy->errorlogIsEnabled() )
    {
      LPTSTR cause = getLastErrorMessage();
      proxy->log(TEXT("error"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("CreatePipe failed (2/2)\n>> %s\n"),
		     cause
		     );
      LocalFree(cause);
    }
    CloseHandle(par_rd);
    CloseHandle(chl_wd_dup);
    if( proxy->tracelogIsEnabled() )
    {
      proxy->log(TEXT("trace"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("failure, CreatePipe (2/2)\n")
		     );
    }
    return FALSE;
  }
  if( !DuplicateHandle(hCurrentProcess, chl_rd,
		       hCurrentProcess, &chl_rd_dup,
		       0,
		       TRUE,
		       DUPLICATE_SAME_ACCESS
		       ) )
  {
    if( proxy->errorlogIsEnabled() )
    {
      LPTSTR cause = getLastErrorMessage();
      proxy->log(TEXT("error"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("DuplicateHandle failed (2/2)\n>> %s\n"),
		     cause
		     );
      LocalFree(cause);
    }
    CloseHandle(par_rd);
    CloseHandle(par_wd);
    CloseHandle(chl_rd);
    CloseHandle(chl_wd_dup);
    if( proxy->tracelogIsEnabled() )
    {
      proxy->log(TEXT("trace"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("failure, DuplicateHandle (2/2)\n")
		     );
    }
    return FALSE;
  }
  CloseHandle(chl_rd);
  chl_rd = NULL;
  
  if( proxy->debuglogIsEnabled() )
  {
    proxy->log(TEXT("debug"),
		   TEXT("RishuProxy::on_load"),
		   TEXT("CreatePipe ok\n")
		   );
  }
  }else if( proxy->communicationType==RP_CT_TCP )
  {
  /* --------------------------------------------------------------------------
   * 通信用のソケットの作成.
   */
  {
    struct sockaddr_in addr;
    int nRes;
    
    /*
    LPTSTR c = getErrorMessage(WSAEINPROGRESS);
    MessageBox(NULL,c,TEXT("Rishu/Proxy"),MB_OK|MB_ICONERROR);
    //MessageBox(NULL,c,MAKEINTRESOURCE(IDS_MBCAPTION),MB_OK|MB_ICONWARNING);
    LocalFree(c);
    */
    
    proxy->sockListen = socket(AF_INET,SOCK_STREAM,0);
    if( proxy->sockListen==INVALID_SOCKET )
    {
      int errcode = WSAGetLastError();
      if( errcode==WSANOTINITIALISED )
      {
	WSADATA wsaData;
	int nRes;
	if( proxy->warnlogIsEnabled() )
        {
	  proxy->log(TEXT("warn"),
			 TEXT("RishuProxy::on_load"),
			 TEXT("[winsock] not initialized yet, now starting up.\n")
			 );
	}
	ZeroMemory(&wsaData,sizeof(wsaData));
	nRes = WSAStartup(MAKEWORD(4,4),&wsaData);
	if( nRes!=0 )
	{
	  /* WSAEINPROGRESS だとしばらく待つべきかも？ */
	  if( proxy->errorlogIsEnabled() )
	  {
	    LPTSTR cause = getErrorMessage(nRes);
	    proxy->log(TEXT("error"),
		       TEXT("RishuProxy::on_load"),
		       TEXT("[WSAStartup, failed]\n"
			    "[[errcode=0d,0x%08x]]\n"
			    "[[\n%s]]\n"),
		       cause
			   );
	    LocalFree(cause);
	  }
	  return FALSE;
	}
	if( proxy->systemlogIsEnabled() )
        {
	  proxy->log(TEXT("system"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("[winsock] starting up, ok.\n"
			  "[[\nVersion=%d.%d\n]]\n"),
		     LOBYTE(wsaData.wVersion),
		     HIBYTE(wsaData.wVersion)
		     );
	}
	if( ((LOBYTE(wsaData.wVersion)<<8)|HIBYTE(wsaData.wVersion)) < 0x0101 )
	{
	  if( proxy->warnlogIsEnabled() )
	  {
	    proxy->log(TEXT("warn"),
		       TEXT("RishuProxy::on_load"),
		       TEXT("maybe winsock so old (%d.%d), v 1.1 and later recomanded.\n"),
		       LOBYTE(wsaData.wVersion),
		       HIBYTE(wsaData.wVersion)
		       );
	  }
	}
	/* も一回つなげてみる. */
	proxy->sockListen = socket(AF_INET,SOCK_STREAM,0);
	if( proxy->sockListen==INVALID_SOCKET )
	{ /* 初期化だけじゃない or やっぱりだめ */
	  if( proxy->errorlogIsEnabled() )
	  {
	    LPTSTR cause = getErrorMessage(WSAGetLastError());
	    proxy->log(TEXT("error"),
			   TEXT("RishuProxy::on_load"),
			   TEXT("[socket, failed]\n>> %s\n"),
			   cause
			   );
	    LocalFree(cause);
	  }
	  return FALSE;
	}
	/* {WSAStartup でsocket 作成に成功} */
      }else
      { /* 未初期化以外のエラー */
	if( proxy->errorlogIsEnabled() )
	{
	  LPTSTR cause = getErrorMessage(WSAGetLastError());
	  proxy->log(TEXT("error"),
			 TEXT("RishuProxy::on_load"),
			 TEXT("[socket, failed]\n>> %s\n"),
			 cause
			 );
	  LocalFree(cause);
	}
	return FALSE;
      }
    } /* socket==INVALID_SOCKET */
    /* ソケット初期化OK */
    
    ZeroMemory(&addr,sizeof(addr));
    addr.sin_family = AF_INET;
    addr.sin_addr.s_addr = htonl(INADDR_LOOPBACK);
    addr.sin_port = htons(0);
    
    nRes = bind(proxy->sockListen,(const struct sockaddr*)&addr,sizeof(addr));
    if( nRes<0 )
    {
      if( proxy->errorlogIsEnabled() )
      {
	int errcode = WSAGetLastError();
	LPTSTR cause = getErrorMessage(errcode);
	proxy->log(TEXT("error"),
		       TEXT("RishuProxy::on_load"),
		       TEXT("[bind, failed]\n[[errcode=%d,0x%08x]]\n[[\n%s]]\n"),
		       errcode, errcode, cause
		       );
	LocalFree(cause);
      }
      closesocket(proxy->sockListen);
      proxy->sockListen = INVALID_SOCKET;
      return FALSE;
    }
    
    nRes = listen(proxy->sockListen,1);
    if( nRes<0 )
    {
      if( proxy->errorlogIsEnabled() )
      {
	LPTSTR cause = getErrorMessage(WSAGetLastError());
	proxy->log(TEXT("error"),
		       TEXT("RishuProxy::on_load"),
		       TEXT("[listen, failed]\n>> %s\n"),
		       cause
		       );
	LocalFree(cause);
      }
      return FALSE;
    }
  }
  
  if( proxy->debuglogIsEnabled() )
  {
    proxy->log(TEXT("debug"),
		   TEXT("RishuProxy::on_load"),
		   TEXT("socket ok\n")
		   );
  }
  }
  
  /* --------------------------------------------------------------------------
   * 実行ファイルの位置.
   * "perl" "rishu_remote.pl" [--port 1234]
   */
  {
    LPCTSTR relative_perl = TEXT("modules\\5.8.0\\bin\\MSWin32-x86-multi-thread\\perl.exe");
    LPCTSTR relative_script = TEXT("rishu_remote.pl");
    DWORD relative_perl_len;
    DWORD relative_script_len;
    
    LPTSTR path_perl;
    LPTSTR path_script;
    DWORD path_perl_len;
    DWORD path_script_len;
    
    /* perl へのパスを変換. */
    relative_perl_len = _tcslen(relative_perl);
    path_perl = proxy->getPath(relative_perl,relative_perl_len);
    if( !path_perl )
    {
      if( proxy->errorlogIsEnabled() )
      {
	proxy->log(TEXT("error"),
		       TEXT("RishuProxy::on_load"),
		       TEXT("proxy->getPath failed (perl.exe)\n")
		       );
      }
      if( proxy->tracelogIsEnabled() )
      {
	proxy->log(TEXT("trace"),
		       TEXT("RishuProxy::on_load"),
		       TEXT("proxy->getPath failed (perl.exe)\n")
		       );
      }
      CloseHandle(err_rd);
      CloseHandle(err_wd_dup);
#ifdef RISHU_PROXY_USEPIPE
      CloseHandle(par_rd);
      CloseHandle(par_wd);
      CloseHandle(chl_rd_dup);
      CloseHandle(chl_wd_dup);
#endif
#ifdef RISHU_PROXY_USESOCKET
      closesocket(proxy->sockListen);
      proxy->sockListen = INVALID_SOCKET;
#endif
      return FALSE;
    }
    
    /* script へのパスを変換. */
    relative_script_len = _tcslen(relative_script);
    path_script = proxy->getPath(relative_script,relative_script_len);
    if( !path_script )
    {
      if( proxy->errorlogIsEnabled() )
      {
	proxy->log(TEXT("error"),
		       TEXT("RishuProxy::on_load"),
		       TEXT("proxy->getPath failed (script)\n")
		       );
      }
      if( proxy->tracelogIsEnabled() )
      {
	proxy->log(TEXT("trace"),
		       TEXT("RishuProxy::on_load"),
		       TEXT("proxy->getPath failed (script)\n")
		       );
      }
      proxy->myFree(path_perl);
      
      CloseHandle(err_rd);
      CloseHandle(err_wd_dup);
#ifdef RISHU_PROXY_USEPIPE
      CloseHandle(par_rd);
      CloseHandle(par_wd);
      CloseHandle(chl_rd_dup);
      CloseHandle(chl_wd_dup);
#endif
#ifdef RISHU_PROXY_USESOCKET
      closesocket(proxy->sockListen);
      proxy->sockListen = INVALID_SOCKET;
#endif
      return FALSE;
    }
    
    /* commandLine に再構成 */
    path_perl_len = proxy->dwLoaddirLen + relative_perl_len;
    path_script_len = proxy->dwLoaddirLen + relative_script_len;
    if( proxy->communicationType==RP_CT_PIPE )
    {
      commandLine = proxy->myAlloc(
				       (path_perl_len+path_script_len+6
					)*sizeof(TCHAR)
				       );
    }else
    {
      commandLine = proxy->myAlloc(
				       (path_perl_len+path_script_len+6+16
					)*sizeof(TCHAR)
				       );
    }
    
    if( !commandLine )
    {
      if( proxy->errorlogIsEnabled() )
      {
	proxy->log(TEXT("error"),
		       TEXT("RishuProxy::on_load"),
		       TEXT("proxy->myAlloc failed (commandLine)\n")
		       );
      }
      if( proxy->tracelogIsEnabled() )
      {
	proxy->log(TEXT("trace"),
		       TEXT("RishuProxy::on_load"),
		       TEXT("proxy->myAlloc failed (commandLine)\n")
		       );
      }
      proxy->myFree(path_perl);
      proxy->myFree(path_script);
      
      CloseHandle(err_rd);
      CloseHandle(err_wd_dup);
#ifdef RISHU_PROXY_USEPIPE
      CloseHandle(par_rd);
      CloseHandle(par_wd);
      CloseHandle(chl_rd_dup);
      CloseHandle(chl_wd_dup);
#endif
#ifdef RISHU_PROXY_USESOCKET
      closesocket(proxy->sockListen);
      proxy->sockListen = INVALID_SOCKET;
#endif
      return FALSE;
    }
    /* commandLine バッファの確保 OK. */
    commandLine[0] = TEXT('\"');
    CopyMemory(commandLine+1,path_perl,path_perl_len*sizeof(TCHAR));
    commandLine[path_perl_len+1] = TEXT('"');
    commandLine[path_perl_len+2] = TEXT(' ');
    commandLine[path_perl_len+3] = TEXT('"');
    CopyMemory(commandLine+path_perl_len+4,path_script,path_script_len);
    commandLine[path_perl_len+4+path_script_len] = TEXT('\"');
    if( proxy->communicationType==RP_CT_PIPE )
    {
      commandLine[path_perl_len+4+path_script_len+1] = TEXT('\0');
    }else
    {
      int nRes;
      int socklen;
      struct sockaddr_in sockaddr;
      socklen = sizeof(sockaddr);
      nRes = getsockname(proxy->sockListen,(struct sockaddr*)&sockaddr,&socklen);
      if( nRes<0 )
      {
	if( proxy->errorlogIsEnabled() )
	{
	  LPTSTR cause = getErrorMessage(WSAGetLastError());
	  proxy->log(TEXT("error"),
			 TEXT("RishuProxy::on_load"),
			 TEXT("getsockname failed\n>> %s\n"),
			 cause
			 );
	  LocalFree(cause);
	}
	if( proxy->tracelogIsEnabled() )
	{
	  LPTSTR cause = getErrorMessage(WSAGetLastError());
	  proxy->log(TEXT("trace"),
			 TEXT("RishuProxy::on_load"),
			 TEXT("getsockname failed\n>> %s\n"),
			 cause
			 );
	  LocalFree(cause);
	}
	proxy->myFree(path_perl);
	proxy->myFree(path_script);
	proxy->myFree(commandLine);
	
	CloseHandle(err_rd);
	CloseHandle(err_wd_dup);
	closesocket(proxy->sockListen);
	proxy->sockListen = INVALID_SOCKET;
	return FALSE;
      }
      nRes = _sntprintf(commandLine+path_perl_len+4+path_script_len+1,
			16,
			TEXT(" --port %u"),
			ntohs(sockaddr.sin_port)
			);
      if( nRes<0 )
      {
	if( proxy->errorlogIsEnabled() )
	{
	  proxy->log(TEXT("error"),
			 TEXT("RishuProxy::on_load"),
			 TEXT("snprintf failed (commandLine)\n")
			 );
	}
	if( proxy->tracelogIsEnabled() )
	{
	  proxy->log(TEXT("trace"),
			 TEXT("RishuProxy::on_load"),
			 TEXT("snprintf failed (commandLine)\n")
			 );
	}
	proxy->myFree(path_perl);
	proxy->myFree(path_script);
	proxy->myFree(commandLine);
	
	CloseHandle(err_rd);
	CloseHandle(err_wd_dup);
	closesocket(proxy->sockListen);
	proxy->sockListen = INVALID_SOCKET;
	return FALSE;
      }
    }
    
    /* 後始末. */
    proxy->myFree(path_perl);
    proxy->myFree(path_script);
  }
  if( proxy->systemlogIsEnabled() )
  {
    proxy->log(TEXT("system"),
		   TEXT("RishuProxy::on_load"),
		   TEXT("[commandLine]\n[[\n%s\n]]\n"),
		   commandLine
		   );
  }

  /* --------------------------------------------------------------------------
   * 子プロセスの作成.
   */
  dwCreationFlags = CREATE_NO_WINDOW;
  
  ZeroMemory(&startupInfo,sizeof(startupInfo));
  startupInfo.cb = sizeof(startupInfo);
#ifndef RISHU_PROXY_USESOCKET
  startupInfo.hStdInput  = chl_rd_dup;
  startupInfo.hStdOutput = chl_wd_dup;
#else
  startupInfo.hStdInput  = NULL;
  startupInfo.hStdOutput = NULL;
#endif
  startupInfo.hStdError  = err_wd_dup;
  startupInfo.dwFlags = STARTF_USESTDHANDLES;
  
  res = CreateProcess(NULL,
		      commandLine,
		      NULL,
		      NULL,
		      TRUE,
		      dwCreationFlags,
		      NULL,
		      NULL,
		      &startupInfo,
		      &processInfo
		      );

  if( !res )
  { /* 子プロセス作成失敗. */
    if( proxy->errorlogIsEnabled() )
    {
      LPTSTR cause = getLastErrorMessage();
      proxy->log(TEXT("error"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("CreateProcess failed\n"
			  ">> %s"
			  "command-line>> %s\n"
			  ),
		     cause,
		     commandLine
		     );
      LocalFree(cause);
    }
    proxy->myFree(commandLine);
    
    CloseHandle(err_rd);
    CloseHandle(err_wd_dup);
#ifndef RISHU_PROXY_USESOCKET
    CloseHandle(par_rd);
    CloseHandle(par_wd);
    CloseHandle(chl_rd_dup);
    CloseHandle(chl_wd_dup);
#else
    closesocket(proxy->sockListen);
    proxy->sockListen = INVALID_SOCKET;
#endif
    if( proxy->tracelogIsEnabled() )
    {
      proxy->log(TEXT("trace"),
		     TEXT("RishuProxy::on_load"),
		     TEXT("failed, CreateProcess\n")
		     );
    }
    return FALSE;
  }
  
  proxy->myFree(commandLine);
  
  proxy->hChlProcess = processInfo.hProcess;
#ifndef RISHU_PROXY_USESOCKET
  proxy->hChlInput = par_wd;
  proxy->hChlOutput = par_rd;
  CloseHandle(chl_rd_dup);
  CloseHandle(chl_wd_dup);
#endif
  proxy->hChlErrorOutput = err_rd;
  CloseHandle(err_wd_dup);
  
  if( proxy->tracelogIsEnabled() )
  {
    proxy->log(TEXT("trace"),
		   TEXT("RishuProxy::on_load"),
		   TEXT("createProcess ok, pid = 0x%#08x\n"),
		   proxy->hChlProcess
		   );
  }
  
#ifdef RISHU_PROXY_USESOCKET
  if( proxy->communicationType==RP_CT_TCP )
  { /* 子プロセスとの接続の確立. */
    fd_set rfds;
    int nRes;
    struct sockaddr_in addr;
    int addrlen;
    unsigned long ulNbio;
    
    FD_ZERO(&rfds);
    FD_SET(proxy->sockListen,&rfds);
    
    nRes = select(proxy->sockListen+1,&rfds,NULL,NULL,&proxy->tvTimeout);
    if( nRes<=0 )
    { /* ERROR or TIMEOUT. */
      if( proxy->errorlogIsEnabled() )
      {
	if( nRes==0 )
        { /* timed out */
	  proxy->log(TEXT("error"),
			 TEXT("RishuProxy::on_load"),
			 TEXT("[select failed (for accept)]\n"
			      "[[\ntimed out (%ld.%06ld secs)\n]]\n"),
			 proxy->tvTimeout.tv_sec,proxy->tvTimeout.tv_usec
			 );
	}else
	{ /* error */
	  LPTSTR cause = getErrorMessage(WSAGetLastError());
	  proxy->log(TEXT("error"),
			 TEXT("RishuProxy::on_load"),
			 TEXT("[select failed (for accept)]\n"
			      "[[\n%s]]\n"),
			 cause
			 );
	  LocalFree(cause);
	}
      }
      closesocket(proxy->sockListen);
      proxy->sockListen = INVALID_SOCKET;
      proxy->logChildStderr();
      return FALSE;
    }
    
    addrlen = sizeof(addr);
    proxy->sockRemote = accept(proxy->sockListen,(struct sockaddr*)&addr,&addrlen);
    if( proxy->sockRemote==INVALID_SOCKET )
    {
      if( proxy->errorlogIsEnabled() )
      {
	LPTSTR cause = getErrorMessage(WSAGetLastError());
	proxy->log(TEXT("error"),
		       TEXT("RishuProxy::on_load"),
		       TEXT("failure, listen()\n>> %s\n"),
		       cause
		       );
	LocalFree(cause);
      }
      return FALSE;
    }
    /* 接続したので Listen ソケットはもういらない */
    closesocket(proxy->sockListen);
    proxy->sockListen = INVALID_SOCKET;
    
    ulNbio = 1;
    nRes = ioctlsocket(proxy->sockRemote,FIONBIO,&ulNbio);
    if( nRes<0 )
    {
      if( proxy->warnlogIsEnabled() )
      {
	LPTSTR cause = getErrorMessage(WSAGetLastError());
	proxy->log(TEXT("warn"),
		       TEXT("RishuProxy::on_load"),
		       TEXT("failure, ioctlsocket(), non-blocking-io\n"
			    "[[\n%s]]\n"),
		       cause
		       );
	LocalFree(cause);
      }
    }
  }
#endif
  
  /* --------------------------------------------------------------------------
   * on_load 通知.
   */
  RishuProxy_Request_init(&request,sizeof(request),0x0101);
  RishuProxy_Request_setPath(&request,"LOAD",4);
  RishuProxy_Request_write(&request,"Directory: ",11);
  RishuProxy_Request_write(&request,(LPCVOID)hGlobal_loaddir,loaddir_len);
  RishuProxy_Request_write(&request,"\r\n\r\n",4);
  RishuProxy_Request_send(&request,proxy);
  RishuProxy_Request_discard(&request);
  
  /* 回答受信 */
  RishuProxy_Response_init(&response,sizeof(response),proxy);
  ptr_response = proxy->recvResponse(&response);
  res = ptr_response && (RishuProxy_Response_getStatusCode(&response)/100)==2;
  if( proxy->tracelogIsEnabled() )
  {
    proxy->log(TEXT("trace"),
		   TEXT("RishuProxy::on_load"),
		   TEXT("[status code]\n[[%d]]\n[response]\n"
			"[[\n%.*s]]\n[[\n%.*s]]\n"),
		   RishuProxy_Response_getStatusCode(&response),
		   response.dwHeaderLen,response.header,
		   response.dwBodyLen,response.body?response.body:""
		   );
  }
  RishuProxy_Response_discard(&response);
  
  if( proxy->tracelogIsEnabled() )
  {
    proxy->log(TEXT("trace"),
		   TEXT("RishuProxy::on_load"),
		   TEXT("leave [result==%s]\n"),
		   res ? "TRUE" : "FALSE"
		   );
    
  }
  return res;
}

/* ----------------------------------------------------------------------------
 * 栞 Method / request
 */
HGLOBAL RishuProxy_on_request(RishuProxy* proxy,
			      HGLOBAL hGlobal,
			      long* ptr_len
			      )
{
  RishuProxy_Request request;
  RishuProxy_Response response;
  RishuProxy_Response* ptr_response;
  
  if( !proxy )
  {
    /* proxy ないので trace もできないし.. */
    return RishuProxy_replyFatal(hGlobal,ptr_len);
  }
  
  if( proxy->tracelogIsEnabled() )
  {
    proxy->log(TEXT("trace"),
		   TEXT("RishuProxy::on_request"),
		   TEXT("[enter] [hGlobal==0x%08x] [len==%d]\n[[\n%.*s]]\n"),
		   hGlobal, *ptr_len,
		   hGlobal?*ptr_len:6, hGlobal?hGlobal:"(null)"
		   );
  }

  if( !proxy->isConnectedForWrite() )
  {
    if( proxy->errorlogIsEnabled() )
    {
      proxy->log(TEXT("error"),
		     TEXT("RishuProxy::on_request"),
		     TEXT("not connected (for write)\n")
		     );
    }
    if( proxy->tracelogIsEnabled() )
    {
      proxy->log(TEXT("trace"),
		     TEXT("RishuProxy::on_request"),
		     TEXT("failure, not connected (for write)\n")
		     );
    }
    hGlobal = RishuProxy_replyFatal(hGlobal,ptr_len);
    if( proxy->tracelogIsEnabled() )
    {
      proxy->log(TEXT("trace"),
		     TEXT("RishuProxy::on_request"),
		     TEXT("[leave] [hGlobal==0x%08x] [len==%d]\n[[\n%.*s]]\n"),
		     hGlobal, *ptr_len,
		     hGlobal?*ptr_len:6, hGlobal?hGlobal:"(null)"
		     );
    }
    return hGlobal;
  }
  
  /* --------------------------------------------------------------------------
   * on_request 通知.
   */
  RishuProxy_Request_init(&request,sizeof(request),0x0101);
  RishuProxy_Request_setPath(&request,"REQUEST",7);
  RishuProxy_Request_write(&request,"Command: ",9);
  RishuProxy_Request_write(&request,(LPCVOID)hGlobal,*ptr_len);
  RishuProxy_Request_write(&request,"\r\n",2);
  RishuProxy_Request_send(&request,proxy);
  RishuProxy_Request_discard(&request);
  
  /* 回答受信 */
  RishuProxy_Response_init(&response,sizeof(response),proxy);
  ptr_response = proxy->recvResponse(&response);
  if( !ptr_response )
  {
    if( proxy->errorlogIsEnabled() )
    {
      proxy->log(TEXT("error"),
		     TEXT("RishuProxy::on_request"),
		     TEXT("no response...\n")
		     );
    }
    RishuProxy_Response_discard(&response);
    
    hGlobal = RishuProxy_reply500(hGlobal,ptr_len);
    
    if( proxy->tracelogIsEnabled() )
    {
      proxy->log(TEXT("trace"),
		     TEXT("RishuProxy::on_request"),
		     TEXT("failure, no response\n")
		     );
    }
    return hGlobal;
  }
  if( (RishuProxy_Response_getStatusCode(&response)/100)!=2 )
  {
    if( proxy->errorlogIsEnabled() )
    {
      proxy->log(TEXT("error"),
		     TEXT("RishuProxy::on_request"),
		     TEXT("got failure [code==%d] [line==%s]\n"),
		     RishuProxy_Response_getStatusCode(&response),
		     RishuProxy_Response_getStatusLine(&response)
		     );
    }
    RishuProxy_Response_discard(&response);
    
    hGlobal = RishuProxy_reply500(hGlobal,ptr_len);
    
    if( proxy->tracelogIsEnabled() )
    {
      proxy->log(TEXT("trace"),
		     TEXT("RishuProxy::on_request"),
		     TEXT("failure, response tells failure\n")
		     );
    }
    return hGlobal;
  }
  
  /* hGlobalに変換. */
  hGlobal = RishuProxy_Response_intoGlobalMemory(&response,hGlobal,ptr_len);
  RishuProxy_Response_discard(&response);
  
  if( proxy->tracelogIsEnabled() )
  {
#if 0
    proxy->log(TEXT("trace"),
		   TEXT("RishuProxy::on_request"),
		   TEXT("[enter] [hGlobal==0x%08x] [len==%d]\n[[\n%.*s]]\n"),
		   hGlobal, *ptr_len,
		   hGlobal?*ptr_len:6, hGlobal?hGlobal:"(null)"
		   );
#else
    BOOL res;
    res = proxy->logOpen(TEXT("trace"),
			 TEXT("RishuProxy::on_request")
			 );
    if( res )
    {
      int i;
      proxy->logText(TEXT("[leave] [hGlobal==0x%08x] [len==%d]\n[[\n"),
		     hGlobal, *ptr_len
		     );
      if( hGlobal==NULL )
      {
	proxy->logText(TEXT("(null)"));
      }else
      {
	const LPBYTE ptr = (LPBYTE)hGlobal;
	int start = 0;
	for( i=0; i<*ptr_len; ++i )
	{
	  if( ptr[i]<0x20 || ptr[i]=='\\' )
	  {
	    if( start!=i )
	    {
	      proxy->logText("%.*s",i-start,ptr+start);
	    }
	    if( ptr[i]=='\n' )
	    {
	      proxy->logText("\\n\n");
	    }else if( ptr[i]=='\r' )
	    {
	      proxy->logText("\\r");
	    }else if( ptr[i]=='\\' )
	    {
	      proxy->logText("\\\\");
	    }else if( ptr[i]=='\0' )
	    {
	      proxy->logText("\\0");
	    }else
	    {
	      proxy->logText("\\x%02x",ptr[i]);
	    }
	    start = i+1;
	  }
	}
	if( start!=i )
	{
	  proxy->logText("%.*s",i-start,ptr+start);
	}
      }
      proxy->logText(TEXT("]]\n"));
      proxy->logClose();
    }
#endif
  }
  
  return hGlobal;
}

/* ----------------------------------------------------------------------------
 * 栞 Method / unload
 */ 
BOOL RishuProxy_on_unload(RishuProxy* proxy)
{
  RishuProxy_Request request;
  RishuProxy_Response response;
  RishuProxy_Response* ptr_response;
  BOOL res;
  
  if( !proxy )
  {
    return TRUE;
  }

  if( proxy->tracelogIsEnabled() )
  {
    proxy->log(TEXT("trace"),
		   TEXT("RishuProxy::on_unload"),
		   TEXT("enter\n")
		   );
  }

  if( !proxy->isConnectedForWrite() )
  {
    if( proxy->errorlogIsEnabled() )
    {
      proxy->log(TEXT("error"),
		     TEXT("RishuProxy::on_unload"),
		     TEXT("not connected (for write)\n")
		     );
    }
    if( proxy->tracelogIsEnabled() )
    {
      proxy->log(TEXT("trace"),
		     TEXT("RishuProxy::on_unload"),
		     TEXT("failure, not connected (for write)\n"
			  "(but results TRUE so unload finished.)")
		     );
    }
    return TRUE;
  }
  
  /* --------------------------------------------------------------------------
   * on_unload 通知.
   */
  RishuProxy_Request_init(&request,sizeof(request),0x0101);
  RishuProxy_Request_setPath(&request,"UNLOAD",6);
  RishuProxy_Request_write(&request,"\r\n",2);
  RishuProxy_Request_send(&request,proxy);
  RishuProxy_Request_discard(&request);
  
  /* 回答受信 */
  RishuProxy_Response_init(&response,sizeof(response),proxy);
  ptr_response = proxy->recvResponse(&response);
  res = ptr_response && (RishuProxy_Response_getStatusCode(&response)/100)==2;
  RishuProxy_Response_discard(&response);
  
  if( proxy->tracelogIsEnabled() )
  {
    proxy->log(TEXT("trace"),
		   TEXT("RishuProxy::on_unload"),
		   TEXT("leave [result==%s]\n"),
		   res ? "TRUE" : "FALSE"
		   );
    
  }
  
  return res;
}

/* ----------------------------------------------------------------------------
 * 子プロセスから受信.
 *   response は必ず一度クリアされます.
 *   エラー時には NULL を返します.
 *   エラー時には, response にエラーが起きるまでのデータは格納されますので 
 *   原因解明に役立つかも？ 
 */
RishuProxy_Response* RishuProxy_recvResponse(RishuProxy* proxy,
					     RishuProxy_Response* response
					     )
{
  LPCBYTE ptr;
  BOOL linefeed;
  DWORD endpos;
  
  /* 先ず response インスタンスのクリア. */
  if( !response )
  {
    if( proxy->errorlogIsEnabled() )
    {
      proxy->log(TEXT("error"),
		     TEXT("RishuProxy::recvResponse"),
		     TEXT("arg:response is required\n")
		     );
    }
    return NULL;
  }
  response->clear();
  
  if( !proxy )
  {
    return NULL;
  }

  /* 接続確認 */
  if( !proxy->isConnectedForRead() )
  {
    if( proxy->errorlogIsEnabled() )
    {
      proxy->log(TEXT("error"),
		     TEXT("RishuProxy::recvResponse"),
		     TEXT("not connected (for read)\n")
		     );
    }
    return NULL;
  }
  
  /* 読み込みバッファの構築 */
  if( !proxy->readbuf )
  {
    proxy->readbuf = proxy->myAlloc(4*1024);
    if( !proxy->readbuf )
    {
      if( proxy->errorlogIsEnabled() )
      {
	proxy->log(TEXT("error"),
		       TEXT("RishuProxy::recvResponse"),
		       TEXT("myAlloc failed\n")
		       );
      }
      return NULL;
    }
    proxy->dwReadBufSize = proxy->mySize(proxy->readbuf);
  }
  
  /* --------------------------------------------------------------------------
   * 読み込み開始.
   */
  ptr = (LPCBYTE)proxy->readbuf;
  linefeed = proxy->dwReadBufLen==0;
  endpos = 0;
  for(;;)
  {
    DWORD i;
    BOOL res;
    if( proxy->tracelogIsEnabled() )
    {
      proxy->log(TEXT("trace"),
		     TEXT("RishuProxy::recvResponse"),
		     TEXT("ReadFile...\n",)
		     );
    }
    res = proxy->async_readFromChild();
    if( !res )
    {
      if( proxy->errorlogIsEnabled() )
      {
	proxy->log(TEXT("error"),
		       TEXT("RishuProxy::recvResponse"),
		       TEXT("[proxy->async_readFromChild, failed]\n")
		       );
      }
      /* 読み込んでた分は追加しちゃう. */
      response->write(ptr,proxy->dwReadBufLen);
      proxy->dwReadBufLen = 0;
      return NULL;
    }

#if 0
    /* 通信内容 */
    for( i=0; i<proxy->dwReadBufLen; ++i )
    {
      register const unsigned char c = ptr[i];
      if( 0x20<=c && c<=0x7e ) putchar(c);
      else if( c=='\r' ) putchar('\\'),putchar('r');
      else if( c=='\n' ) putchar('\\'),putchar('n');
      else printf("\\x%02x",c);
    }
    putchar('\n');
#endif
    
    for( i=0; i<proxy->dwReadBufLen; ++i )
    {
      while( ptr[i]=='\r' && i+1<proxy->dwReadBufLen )
      {
	if( proxy->debuglogIsEnabled() )
	{
	  proxy->log(TEXT("debug"),
			 TEXT("RishuProxy::recvResponse"),
			 TEXT("(%3d) skip \\r\n"),
			 i
			 );
	}
	++i;
      }
      if( ptr[i]=='\n' )
      {
	if( linefeed )
	{ /* 空行の発見. */
	  if( proxy->debuglogIsEnabled() )
	  {
	    proxy->log(TEXT("debug"),
			   TEXT("RishuProxy::recvResponse"),
			   TEXT("(%3d) end-of-header found\n"),
			   i
			   );
	  }
	  endpos = i+1;
	  break;
	}else
	{ /* 行末. */
	  linefeed = TRUE;
	  if( proxy->debuglogIsEnabled() )
	  {
	    proxy->log(TEXT("debug"),
			   TEXT("RishuProxy::recvResponse"),
			   TEXT("(%3d) line feed found\n"),
			   i
			   );
	  }
	}
      }else if( linefeed )
      {
	if( proxy->debuglogIsEnabled() )
	{
	  proxy->log(TEXT("debug"),
			 TEXT("RishuProxy::recvResponse"),
			 TEXT("(%3d) cancel line feed\n"),
			 i
			 );
	}
	linefeed = FALSE;
      }
    }
    if( endpos )
    {
      response->write(ptr,endpos);
      proxy->dwReadBufLen -= endpos;
      if( proxy->dwReadBufLen )
      {
	MoveMemory(proxy->readbuf,ptr+endpos,proxy->dwReadBufLen);
      }
      break;
    }
    if( proxy->dwReadBufLen>2 )
    {
      response->write(ptr,proxy->dwReadBufLen-1);
      *(LPBYTE)proxy->readbuf = ptr[proxy->dwReadBufLen-1];
      proxy->dwReadBufLen = 1;
    }
    linefeed = FALSE;
  }
  /* ヘッダの読みとりおわり. */
  {
    /* Content-Length: のチェック. */
    LPCBYTE ptr = response->header;
    unsigned long contentLength = 0;
    const LPCBYTE tail = (LPCBYTE)response->header+response->dwHeaderLen-16;
    while( ptr<tail )
    {
      if( strnicmp(ptr,"Content-Length:",15)==0 )
      {
	char* endptr;
	ptr += 16;
	contentLength = strtoul(ptr,&endptr,0);
	if( proxy->debuglogIsEnabled() )
	{
	  proxy->log(TEXT("debug"),
			 TEXT("RishuProxy::recvResponse"),
			 TEXT("[found content-length] [[%.*s]]\n"),
			 endptr-ptr,ptr
			 );
	}
	break;
      }
      if( proxy->debuglogIsEnabled() )
      {
	proxy->log(TEXT("debug"),
		       TEXT("RishuProxy::recvResponse"),
		       TEXT("[skip not content-length header] [[pos=%d,%.*s]]\n"),
		       ptr-(LPBYTE)response->header,
		       15, ptr
		       );
      }
      while( *ptr++!='\n' && ptr<tail )
      {
      }
    }
    if( contentLength>0 )
    { /* Content-Length: 分読み込み. */
      DWORD ateLen;
      
      if( proxy->tracelogIsEnabled() )
      {
	proxy->log(TEXT("trace"),
		       TEXT("RishuProxy::recvResponse"),
		       TEXT("[content-length] [[%d]]\n"),
		       contentLength
		       );
      }
      /* 必要な分のバッファを確保. */
      if( response->dwBodySize<contentLength )
      {
	proxy->lock();
	if( response->body )
	{
	  response->body = proxy->async_myReAlloc(response->body,contentLength);
	}else
	{
	  response->body = proxy->async_myAlloc(contentLength);
	}
	if( !response->body )
	{
	  proxy->unlock();
	  if( proxy->tracelogIsEnabled() )
	  {
	    proxy->log(TEXT("error"),
			   TEXT("RishuProxy::recvResponse"),
			   TEXT("[proxy->async_myAlloc failed]\n")
			   );
	  }
	  response->dwBodyLen = 0;
	  response->dwBodySize = 0;
	  return NULL;
	}
	response->dwBodySize = proxy->async_mySize(response->body);
	proxy->unlock();
      }
      /* body の読み込み〜. */
      ateLen = 0;
      for(;;)
      {
	BOOL res;
	
	if( proxy->dwReadBufLen>=contentLength-ateLen )
	{ /* バッファ内に十分ある♪ */
	  DWORD eatLen = contentLength-ateLen;
	  CopyMemory((LPBYTE)response->body+ateLen,
		     proxy->readbuf,
		     eatLen
		     );
	  proxy->dwReadBufLen -= eatLen;
	  response->dwBodyLen = contentLength;
	  if( proxy->dwReadBufLen )
	  {
	    MoveMemory(proxy->readbuf,(LPCBYTE)proxy->readbuf+eatLen,proxy->dwReadBufLen);
	  }
	  break;
	}
	/* あるだけこぴー */
	CopyMemory((LPBYTE)response->body+ateLen,
		   proxy->readbuf,
		   proxy->dwReadBufLen
		   );
	ateLen += proxy->dwReadBufLen;
	proxy->dwReadBufLen = 0;
	/* もっと読み込み〜 */
	res = proxy->async_readFromChild();
	if( !res )
	{
	  if( proxy->errorlogIsEnabled() )
	  {
	    proxy->log(TEXT("error"),
			   TEXT("RishuProxy::recvResponse"),
			   TEXT("[proxy->async_readFromChild, failed]\n")
			   );
	  }
	  /* データ長はここまでにしておいてみたり. */
	  response->dwBodyLen = ateLen;
	  return NULL;
	}
      }
      
    } /* contentLength>0 */
  }
  if( proxy->debuglogIsEnabled() )
  {
    proxy->log(TEXT("debug"),
		   TEXT("RishuProxy::recvResponse"),
		   TEXT("[succ]\n"
			"[content-length=%d]\n"
			"[content]\n"
			"[[\n%.*s]]\n"),
		   response->dwBodyLen,
		   response->dwBodyLen, response->body
		   );
  }
  
  return response;
}

/* ----------------------------------------------------------------------------
 * 子プロセスと通信 - read
 */
BOOL RishuProxy_async_readFromChild(RishuProxy* proxy)
{
  if( !proxy )
  {
    return FALSE;
  }
  
  if( !proxy->isConnectedForRead() )
  {
    if( proxy->errorlogIsEnabled() )
    {
      proxy->log(TEXT("error"),
		     TEXT("RishuProxy::async_readFromChild"),
		     TEXT("not connected (for read)\n")
		     );
    }
    return FALSE;
  }

#ifndef RISHU_PROXY_USESOCKET
  {
    BOOL res;
    DWORD readbytes;
  
    res = ReadFile(proxy->hChlOutput,
		   (LPBYTE)proxy->readbuf+proxy->dwReadBufLen,
		   proxy->dwReadBufSize-proxy->dwReadBufLen,
		   &readbytes,
		   NULL
		   );
    if( res )
    {
      if( proxy->tracelogIsEnabled() )
      {
	proxy->log(TEXT("trace"),
		       TEXT("RishuProxy::async_readFromChild"),
		       TEXT("[ReadFile]\n[[\n%.*s]]\n",),
		       readbytes,(LPCBYTE)proxy->readbuf+proxy->dwReadBufLen
		       );
      }
      proxy->dwReadBufLen += readbytes;
      return res;
    }
    if( !res )
    {
      if( proxy->errorlogIsEnabled() )
      {
	LPTSTR cause = getLastErrorMessage();
	if( GetLastError()==ERROR_HANDLE_EOF
	    || GetLastError()==ERROR_BROKEN_PIPE )
        {
	  proxy->log(TEXT("error"),
			 TEXT("RishuProxy::async_readFromChild"),
			 TEXT("[ReadFile] reach End-Of-File\n"
			      "[[\n%s]]\n"),
			 cause
			 );
	}else
	{
	  proxy->log(TEXT("error"),
			 TEXT("RishuProxy::async_readFromChild"),
			 TEXT("[ReadFile] error code [%08x]\n"
			      "[[\n%s]]\n"),
			 GetLastError(),
			 cause
			 );
        }
	LocalFree(cause);
      }
    }
    return res;
  }
#else
  {
    fd_set rfds;
    int nRes;
    
    FD_ZERO(&rfds);
    FD_SET(proxy->sockRemote,&rfds);
    nRes = select(proxy->sockRemote+1,&rfds,NULL,NULL,&proxy->tvTimeout);
    //nRes=1;FD_SET(proxy->sockRemote,&rfds);
    if( nRes<=0 )
    { /* ERROR or TIMEOUT. */
      if( proxy->errorlogIsEnabled() )
      {
	if( nRes==0 )
        { /* timed out */
	  proxy->log(TEXT("error"),
			 TEXT("RishuProxy::async_readFromChild"),
			 TEXT("[select failed]\n"
			      "[[\ntimed out (%ld.%06ld secs)\n]]\n"),
			 proxy->tvTimeout.tv_sec,proxy->tvTimeout.tv_usec
			 );
	}else
	{ /* error */
	  LPTSTR cause = getErrorMessage(WSAGetLastError());
	  proxy->log(TEXT("error"),
			 TEXT("RishuProxy::async_readFromChild"),
			 TEXT("[select failed]\n"
			      "[[\n%s]]\n"),
			 cause
			 );
	  LocalFree(cause);
	}
      }
      closesocket(proxy->sockRemote);
      proxy->sockRemote = INVALID_SOCKET;
      proxy->logChildStderr();
      return FALSE;
    }
    if( FD_ISSET(proxy->sockRemote,&rfds) )
    {
      nRes = recv(proxy->sockRemote,
		  (LPBYTE)proxy->readbuf+proxy->dwReadBufLen,
		  proxy->dwReadBufSize-proxy->dwReadBufLen,
		  0
		  );
      if( nRes>0 )
      {
	if( proxy->tracelogIsEnabled() )
        {
	  proxy->log(TEXT("trace"),
			 TEXT("RishuProxy::async_readFromChild"),
			 TEXT("[recv]\n[[\n%.*s]]\n",),
			 nRes,(LPCBYTE)proxy->readbuf+proxy->dwReadBufLen
			 );
	}
	proxy->dwReadBufLen += nRes;
	return TRUE;
      }else
      {
	if( proxy->errorlogIsEnabled() )
        {
	  if( nRes==0 )
	  { /* grasefully close */
	    proxy->log(TEXT("error"),
			   TEXT("RishuProxy::async_readFromChild"),
			   TEXT("[recv, reach End-Of-File]\n"
				"[[\ndetect  gracefully close.\n]]\n")
			   );
	  }else
	  { /* error */
	    LPTSTR cause = getErrorMessage(WSAGetLastError());
	    proxy->log(TEXT("error"),
			   TEXT("RishuProxy::async_readFromChild"),
			   TEXT("[recv, failed]\n"
				"[[\n%s]]\n"),
			   cause
			   );
	    LocalFree(cause);
	  }
	}
	closesocket(proxy->sockRemote);
	proxy->sockRemote = INVALID_SOCKET;
	proxy->logChildStderr();
	return FALSE;
      }
    }
    
    if( proxy->errorlogIsEnabled() )
    {
      proxy->log(TEXT("error"),
		     TEXT("RishuProxy::async_readFromChild"),
		     TEXT("[REACH NOT REACHABLE]\n")
		     );
    }
    
    return FALSE;
  }
#endif
}

/* ----------------------------------------------------------------------------
 * 子プロセスと通信 - write
 */
BOOL RishuProxy_async_writeToChild(RishuProxy* proxy,
				   LPCVOID buf,
				   DWORD  buflen,
				   DWORD* ptr_wrotelen
				   )
{
  if( !proxy )
  {
    return FALSE;
  }
  
  if( !proxy->isConnectedForWrite() )
  {
    if( proxy->errorlogIsEnabled() )
    {
      proxy->log(TEXT("error"),
		     TEXT("RishuProxy::async_writeToChild"),
		     TEXT("not connected (for write)\n")
		     );
    }
    return FALSE;
  }

#ifndef RISHU_PROXY_USESOCKET
  {
    BOOL res;
    res = WriteFile(proxy->hChlInput,
		    buf,
		    len,
		    ptr_wrotelen,
		    NULL
		    );
    if( !res )
    {
      if( proxy->errorlogIsEnabled() )
      {
	LPTSTR cause = getLastErrorMessage();
	proxy->log(TEXT("error"),
		       TEXT("RishuProxy::async_writeToChild"),
		       TEXT("[WriteFile failed]\n[[\n%s]]\n"),
		       cause
		       );
	LocalFree(cause);
      }
      return FALSE;
    }
  }
  return TRUE;
#else
  {
    fd_set wfds;
    int nRes;
    
    *ptr_wrotelen = 0;
    FD_ZERO(&wfds);
    FD_SET(proxy->sockRemote,&wfds);
    for(;;)
    {
      nRes = select(proxy->sockRemote+1,NULL,&wfds,NULL,&proxy->tvTimeout);
      if( nRes<=0 )
      { /* ERROR or TIMEOUT. */
	if( proxy->errorlogIsEnabled() )
        {
	  if( nRes==0 )
          { /* timed out */
	    proxy->log(TEXT("error"),
			   TEXT("RishuProxy::async_writeToChild"),
			   TEXT("[select failed]\n"
				"[[\ntimed out (%ld.%06ld secs)\n]]\n"),
			   proxy->tvTimeout.tv_sec,proxy->tvTimeout.tv_usec
			   );
	  }else
	  { /* error */
	    LPTSTR cause = getErrorMessage(WSAGetLastError());
	    proxy->log(TEXT("error"),
			   TEXT("RishuProxy::async_writeToChild"),
			   TEXT("[select failed]\n"
				"[[\n%s]]\n"),
			   cause
			   );
	    LocalFree(cause);
	  }
	}
	closesocket(proxy->sockRemote);
	proxy->sockRemote = INVALID_SOCKET;
	return FALSE;
      }
      
      nRes = send(proxy->sockRemote,
		  buf,
		  buflen,
		  0
		  );
      if( nRes>=0 )
      {
	if( proxy->tracelogIsEnabled() )
        {
	  proxy->log(TEXT("trace"),
			 TEXT("RishuProxy::async_sendToChild"),
			 TEXT("[send]\n"
			      "[[wrotelen=%d/%d]]\n"
			      "[[\n%.*s]]\n"),
			 nRes, buflen+*ptr_wrotelen,
			 nRes, buf
			 );
	}
	*ptr_wrotelen += nRes;
	if( nRes!=buflen )
	{ /* まだ一部だけ. */
	  buflen -= nRes;
	  buf = (LPCBYTE)buf+nRes;
	  continue;
	}
	return TRUE;
      }else
      {
	const int errcode = WSAGetLastError();
	if( errcode==WSAEWOULDBLOCK )
	{
	  continue;
	}
	if( proxy->errorlogIsEnabled() )
        {
	  LPTSTR cause = getErrorMessage(errcode);
	  proxy->log(TEXT("error"),
			 TEXT("RishuProxy::async_sendToChild"),
			 TEXT("[send, failed]\n"
			      "[[\n%s]]\n"),
			 cause
			 );
	  LocalFree(cause);
	}
	closesocket(proxy->sockRemote);
	proxy->sockRemote = INVALID_SOCKET;
	return FALSE;
      }
      break;
    }
    
    if( proxy->errorlogIsEnabled() )
    {
      proxy->log(TEXT("error"),
		     TEXT("RishuProxy::async_sendToChild"),
		     TEXT("[REACH NOT REACHABLE]\n")
		     );
    }
    
    return FALSE;
  }
#endif
}

/* ----------------------------------------------------------------------------
 * 子プロセスと通信 - write flush
 */
BOOL RishuProxy_async_flushWriteToChild(RishuProxy* proxy)
{
#ifndef RISHU_PROXY_USESOCKET
  BOOL res;
  res = FlushFileBuffers(proxy->hChlInput);
  if( res )
  {
    return TRUE;
  }
  if( proxy->warnlogIsEnabled() )
  {
    LPTSTR cause = getLastErrorMessage();
    proxy->log(TEXT("warn"),
		   TEXT("RishuProxy::async_flushWriteToChild"),
		   TEXT("[FlushFileBuffers failed]\n"
			"[[\n%s]]\n"),
		   cause
		   );
    LocalFree(cause);
  }
  return FALSE;
#else
  /* ソケットに FlushFileBuffers はできない. */
  return TRUE;
#endif
}

/* ----------------------------------------------------------------------------
 * 標準エラーをログ.
 */
void RishuProxy_logChildStderr(RishuProxy* proxy)
{
  BOOL bLogOpened;
  
  if( !proxy )
  {
    return;
  }
  if( !proxy->hChlErrorOutput )
  {
    return;
  }
  
  bLogOpened = FALSE;
  for(;;)
  {
    DWORD nRes;
    
    /* 状態チェックしてすぐ返す. */
    nRes = WaitForSingleObject(proxy->hChlErrorOutput,0);
    
    if( nRes==WAIT_OBJECT_0 )
    { /* 読み込み可能. */
      BOOL res;
      DWORD readbytes;
      
      break;
      res = ReadFile(proxy->hChlErrorOutput,
		     (LPBYTE)proxy->readbuf+proxy->dwReadBufLen,
		     proxy->dwReadBufSize-proxy->dwReadBufLen,
		     &readbytes,
		     NULL
		     );
      if( res )
      {
	if( !bLogOpened )
	{
	  bLogOpened = proxy->logOpen(TEXT("stderr"),
					  TEXT("RishuProxy::logChildStderr")
					  );
	  if( !bLogOpened )
	  {
	    continue;
	  }
	}
	proxy->logText(
			   TEXT("[ReadFile] [[len==%d]]\n[[\n%.*s]]\n"),
			   readbytes,
			   readbytes,
			   (LPCBYTE)proxy->readbuf+proxy->dwReadBufLen
			   );
	continue;
      }
      /* ReadFile 失敗. */
      if( proxy->errorlogIsEnabled() )
      {
	const DWORD errcode = GetLastError();
	LPTSTR cause = getErrorMessage(errcode);
	if( bLogOpened )
	{
	  proxy->logClose();
	  bLogOpened = FALSE;
	}
	if( errcode==ERROR_HANDLE_EOF
	    || errcode==ERROR_BROKEN_PIPE )
	{
	  proxy->log(TEXT("error"),
			 TEXT("RishuProxy::logChildStderr"),
			 TEXT("[ReadFile] reach End-Of-File\n"
			      "[[\n%s]]\n"),
			 cause
			 );
	}else
	{
	  proxy->log(TEXT("error"),
			 TEXT("RishuProxy::logChildStderr"),
			 TEXT("[ReadFile] error code [%08x]\n"
			      "[[\n%s]]\n"),
			 errcode,
			 cause
			 );
	}
	LocalFree(cause);
      }
      return;
    }
    
    /* タイムアウトとか失敗とか. */
    if( nRes==WAIT_TIMEOUT )
    {
      if( bLogOpened )
      {
	proxy->logClose();
	bLogOpened = FALSE;
      }
      return;
    }else if( nRes==WAIT_FAILED )
    {
      if( proxy->errorlogIsEnabled() )
      {
        LPTSTR cause = getLastErrorMessage();
	if( bLogOpened )
        {
	  proxy->logClose();
	  bLogOpened = FALSE;
	}
	proxy->log(TEXT("warn"),
		       TEXT("RishuProxy::logChildStderror"),
		       TEXT("[WaitForSingleObject failed]\n"
			    "[[\n%s]]\n"),
		       cause
		       );
	LocalFree(cause);
      }
    }else if( bLogOpened )
    {
      proxy->logClose();
    }
    break;
  }
  
  return;
}

/* ----------------------------------------------------------------------------
 * ログ (一般型) 
 */
void RishuProxy_log(RishuProxy* proxy,
		    LPCTSTR name,
		    LPCTSTR title,
		    LPCTSTR format,
		    ...
		    )
{
  va_list ap;
  va_start(ap,format);
  proxy->logv(name,title,format,ap);
  va_end(ap);
}

/* ----------------------------------------------------------------------------
 * ログ (va_list 型) 
 */
void RishuProxy_logv(RishuProxy* proxy,
		     LPCTSTR name,
		     LPCTSTR title,
		     LPCTSTR format,
		     va_list ap
		     )
{
  if( proxy->logOpen(name,title) )
  {
    proxy->logTextv(format,ap);
    proxy->logClose();
  }
}

/* ----------------------------------------------------------------------------
 * ログの始まり.
 */
BOOL RishuProxy_logOpen(RishuProxy* proxy,
			LPCTSTR name,
			LPCTSTR title
			)
{
  if( !proxy )
  {
    return FALSE;
  }
  
  /* ログのふるい分け */
  if( _tcscmp(name,TEXT("fatal"))==0 )
  {
    if( !proxy->fatallogIsEnabled() )
    {
      return FALSE;
    }
  }else if( _tcscmp(name,TEXT("error"))==0 )
  {
    if( !proxy->errorlogIsEnabled() )
    {
      return FALSE;
    }
  }else if( _tcscmp(name,TEXT("warn"))==0 )
  {
    if( !proxy->warnlogIsEnabled() )
    {
      return FALSE;
    }
  }else if( _tcscmp(name,TEXT("system"))==0 )
  {
    if( !proxy->systemlogIsEnabled() )
    {
      return FALSE;
    }
  }else if( _tcscmp(name,TEXT("trace"))==0 )
  {
    if( !proxy->tracelogIsEnabled() )
    {
      return FALSE;
    }
  }else if( _tcscmp(name,TEXT("debug"))==0 )
  {
    if( !proxy->debuglogIsEnabled() )
    {
      return FALSE;
    }
  }else if( _tcscmp(name,TEXT("stderr"))==0 )
  {
    if( !proxy->stderrlogIsEnabled() )
    {
      return FALSE;
    }
  }else
  {
    return FALSE;
  }
  
  {
    /* ログ出力 */
    SYSTEMTIME systemTime;
    FILETIME fileTime;
    FILETIME localFileTime;
    DWORD tinysec;
    BOOL res;
    
    if( proxy->dwLogBits & (1<<RishuProxy_LOG_CLEANUP) )
    {
      /* 既存があるかチェック. */
      FILE* fp;
      fp = _tfopen(proxy->logfilename,TEXT("rb"));
      if( fp )
      {
	LPTSTR bakfilename;
	DWORD len;
	BOOL res;
	
	fclose(fp);
	len = _tcslen(proxy->logfilename);
	bakfilename = (LPTSTR)proxy->myAlloc((len+1)*sizeof(TCHAR));
	if( !bakfilename )
	{
	  return FALSE;
	}
	CopyMemory(bakfilename,proxy->logfilename,(len-3)*sizeof(TCHAR));
	CopyMemory(bakfilename+len-3,TEXT("bak"),4*sizeof(TCHAR));
	res = MoveFile(proxy->logfilename,bakfilename);
	proxy->myFree(bakfilename);
      }
      proxy->dwLogBits &= ~(1<<RishuProxy_LOG_CLEANUP);
      if( !proxy->fp )
      {
	proxy->fp = _tfopen(proxy->logfilename,TEXT("wb"));
	if( !proxy->fp )
	{
	  return FALSE;
	}
      }
    }else
    {
      if( !proxy->fp )
      {
	proxy->fp = _tfopen(proxy->logfilename,TEXT("ab"));
	if( !proxy->fp )
	{
	  return FALSE;
	}
      }
    }
    
    /* 時間の取得 */
    GetSystemTimeAsFileTime(&fileTime);
    res = FileTimeToLocalFileTime(&fileTime,&localFileTime)
      && FileTimeToSystemTime(&localFileTime,&systemTime);
    if( !res )
    {
      GetLocalTime(&systemTime);
    }
    /* 100 nanosec == 1/10 usec */
    tinysec = (fileTime.dwLowDateTime)%10000000;
    
    /* 出力. */
    _ftprintf(proxy->fp,
	      TEXT("<< %s (%s) >>\n"
		   "@ %04d/%02d/%02d %02d:%02d:%02d.%06d @\n"
		      ),
	      title, name,
	      systemTime.wYear, systemTime.wMonth, systemTime.wDay,
	      systemTime.wHour, systemTime.wMinute, systemTime.wSecond,
	      tinysec/10
	      );
  }
  return TRUE;
}

/* ----------------------------------------------------------------------------
 * ログのテキスト
 */
BOOL RishuProxy_logText(RishuProxy* proxy,
			LPCTSTR format,
			...
			)
{
  BOOL res;
  va_list ap;
  va_start(ap,format);
  res = proxy->logTextv(format,ap);
  va_end(ap);
  return res;
}
  
/* ----------------------------------------------------------------------------
 * ログのテキスト (va_list 型)
 */
BOOL RishuProxy_logTextv(RishuProxy* proxy,
			 LPCTSTR format,
			 va_list ap
			 )
{
  return _vftprintf(proxy->fp,format,ap);
}

/* ----------------------------------------------------------------------------
 * ログの末尾
 */
void RishuProxy_logClose(RishuProxy* proxy)
{
  _ftprintf(proxy->fp,TEXT("--------\n"));
  fflush(proxy->fp);
}

/* ----------------------------------------------------------------------------
 * makeReply
 *   回答の作成.
 */
HGLOBAL RishuProxy_makeReply(HGLOBAL hGlobal,
			     long* ptr_len,
			     LPCVOID data,
			     DWORD datalen
			     )
{
  if( (DWORD)*ptr_len < datalen && GlobalSize(hGlobal)<datalen )
  {
    hGlobal = GlobalReAlloc(hGlobal,datalen,0);
    if( !hGlobal )
    {
      *ptr_len = 0;
      return NULL;
    }
  }
  CopyMemory((LPVOID)hGlobal,data,datalen);
  *ptr_len = datalen;
  return hGlobal;
}

/* ----------------------------------------------------------------------------
 * reply500
 *   500 Internal Server Error な回答の作成.
 */
HGLOBAL RishuProxy_reply500(HGLOBAL hGlobal,
			    long* ptr_len
			    )
{
  const char* msg = 
    "SHIORI/3.0 500 Internal Server Error \r\n"
    "Sender: Rishu/Proxy\r\n"
    "\r\n";
  return RishuProxy_makeReply(hGlobal,ptr_len,msg,strlen(msg));
}

/* ----------------------------------------------------------------------------
 * replyFatal
 * 異常時の終了メッセージの作成
 *   500 Internal Server Error を返す.
 *   でも, ID: OnSecondChange 時には終了メッセージを返す.
 */
HGLOBAL RishuProxy_replyFatal(HGLOBAL hGlobal,
			      long* ptr_len
			      )
{
  const char* ptr;
  int i;
  BOOL bLinefeed;
  
  bLinefeed = TRUE;
  ptr =  (const char*)hGlobal;
  for( i=0; i<*ptr_len; ++i )
  {
    if( !bLinefeed )
    {
      if( ptr[i]=='\n' )
      {
	bLinefeed = TRUE;
      }
      continue;
    }
    bLinefeed = FALSE;
    if( *ptr_len-i<18 )
    {
      continue;
    }
    if( strnicmp(ptr+i,"ID:",3)!=0 )
    {
      continue;
    }
    i += 3;
    while( isspace(ptr[i]) )
    {
      ++i;
    }
    
    if( *ptr_len-i<14 || strnicmp(ptr+i,"OnSecondChange",14)!=0 )
    {
      continue;
    }
    i+=14;
    if( i<*ptr_len && !isalnum(ptr[i]) )
    {
      const char* msg =
	"SHIORI/3.0 200 OK\r\n"
	/* "Value: 栞システムが異常終了しました。\\x\\-\r\n" */
	"Value: \xe6\xa0\x9e\xe3\x82\xb7\xe3\x82\xb9\xe3\x83\x86\xe3\x83\xa0\xe3\x81\x8c\xe7\x95\xb0\xe5\xb8\xb8\xe7\xb5\x82\xe4\xba\x86\xe3\x81\x97\xe3\x81\xbe\xe3\x81\x97\xe3\x81\x9f\xe3\x80\x82\\x\\-\r\n"
	"Charset: UTF-8\r\n"
	"\r\n";
      return RishuProxy_makeReply(hGlobal,ptr_len,msg,strlen(msg));
    }
  }
  {
  const char* msg = 
    "SHIORI/3.0 500 Internal Server Error \r\n"
    "Sender: Rishu/Proxy\r\n"
    "\r\n";
  return RishuProxy_makeReply(hGlobal,ptr_len,msg,strlen(msg));
  }
}

