## ----------------------------------------------------------------------------
# rishu_remote.pl
#   里珠・外部のプロセス側.
# -----------------------------------------------------------------------------
# Rishu::Remote->new()->run() を呼ぶと, あとは run の中から適宜
# main::on_load, etc.. が呼ばれます.
# 
# -----------------------------------------------------------------------------
# Mastering programed by YAMASHINA Hio
#
# Copyright 2002 YMIRLINK,Inc.
# -----------------------------------------------------------------------------
# $Id: rishu_remote.pl,v 1.11 2002/12/21 09:28:59 hio Exp $
# -----------------------------------------------------------------------------
package main;
use strict;

# -----------------------------------------------------------------------------
# Rishu::Remote 初期化.

#   lib/ ディレクトリを @INC に追加.
BEGIN
{
  my $here = $0;
  if( $^O eq 'MSWin32' )
  {
    $here =~ s/[^\\]+$//;
    $here or $here = '.\\'
  }else
  {
    $here =~ s,[^/]+$,,;
    $here or $here = './'
  }
  unshift @INC,$here.'lib';
}

# instance の構築.
use Rishu::Remote;
Rishu::Remote->instance(@ARGV);

# 里珠の初期化, ここまで.
# -----------------------------------------------------------------------------

# -----------------------------------------------------------------------------
# 里珠つかう人の初期化

# このあたりにわきゃわきゃと.



# -----------------------------------------------------------------------------
# Rishu::Remote 起動.
#   run() 呼び出しで, 終了までブロックします.

Rishu::Remote->instance()->run();

# -----------------------------------------------------------------------------
# 以下, 里珠ハンドラ

# -----------------------------------------------------------------------------
# on_load
sub on_load
{
  my $data = shift;
  $data;
}

# -----------------------------------------------------------------------------
# on_request
sub on_request
{
  my $data = shift;
  if( $data =~ /^thread/i )
  {
    thread_test($data);
  }elsif( $data =~ /^fork/i )
  {
    fork_test($data);
  }elsif( $data =~ /^exit/i )
  {
    exit_test($data);
  }else
  {
    "SHIORI/3.0 200 ok\r\n".
    "Message: enter 'thread', 'fork', or 'exit', and empty line.\r\n".
    "Repeat-of-Request: ".$data;
  }
}

my @shared : shared;
sub thread_test
{
  my $data = shift;
  use threads;
  use threads::shared;
  foreach my $n (1..3)
  {
    threads->new(sub{
      {
	sleep($n);
	lock(@shared);
	push(@shared,threads->self()->tid());
      }
    })->detach();
  }
  my $res = threads->new(sub{
    threads->self()->tid();
  })->join();
  my $fin;
  {
    lock(@shared);
    $fin = join(',',@shared);
    @shared = ();
  }
  "SHIORI/3.0 200 Thread ok\r\n".
  "Join-Thread-Id: $res\r\n".
  "Finished-Detached-Thread-Id: $fin".
  "Repeat-of-Request: ".$data;
}

sub fork_test
{
  my $data = shift;
  my $pid = fork();
  if( !defined($pid) )
  {
    "SHIORI/3.0 500 Fork failed\r\n".
    "Error-Message: $!\r\n".
    "Repeat-of-Request: ".$data;
  }elsif( $pid )
  { # 親プロセス
    my $pid_got = waitpid(-1,0);
    "SHIORI/3.0 200 Fork ok\r\n".
    "Child-Pid: $pid_got\r\n".
    "Repeat-of-Request: ".$data;
  }else
  { # 子プロセス.
    sleep(1);
    exit(0);
  }
}

sub exit_test
{
  my $data = shift;
  Rishu::Remote->instance()->log('trace',"exit\n");
  exit 0;
}

# -----------------------------------------------------------------------------
# on_unload
sub on_unload
{
  1;
}

1;
